/*
# Copyright (C) 2002 by Balazs Nagy <js@iksz.hu>.  All rights reserved.
#
# This program is a free software; you can redistribute it and/or modify
# it under the same terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
*/

/*

   Imprisoning commands

   Steps:

   1| chroot
   1.1| chroot to dir
   1.2| chdir to root

   2| capabilities
   2.1| set capabilities (plus cap_setuid,cap_setgid+ep)
   2.2| set keepcaps (linux feature)

   3| users
   3.1| initgroups
   3.2| setgid
   3.3| setuid

   4| exec

*/

#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/capability.h>
#include <sys/prctl.h>

#include "version.h"

void print_usage(void)
{
	fputs("Usage: imprison [--root new_rootdir] [--caps kept_capabilities]\n", stderr);
	fputs("                [--user set_username] [--set-groups] command [params]\n",stderr);
	fputs("or:    imprison [-r new_rootdir] [-c kept_capabilities]\n", stderr);
	fputs("                [-u set_username] [-G] command [params]\n", stderr);
}

void die(const char *str)
{
	fputs(str, stderr);
	if (errno) {
		fputs(": ", stderr);
		fputs(strerror(errno), stderr);
	}
	fputs("\n", stderr);
	print_usage();
	exit(1);
}

void err_die(const char *str)
{
	fputs("imprison fatal: ", stderr);
	die(str);
}

void err_die2(const char *str, const char *str2)
{
	fputs("imprison fatal: ", stderr);
	fputs(str, stderr);
	die(str2);
}

void help_die(void)
{
	fputs(PKGNAME, stderr);
	fputs(" v", stderr);
	fputs(VERSION, stderr);
	fputs("\n\n", stderr);
	print_usage();
	fputs("  where\n", stderr);
	fputs("    --root or -r: chroots to directory\n", stderr);
	fputs("    --caps or -c: set capabilities in cap_from_text(3) format\n", stderr);
	fputs("    --user or -u: set uid and gid to username\n", stderr);
	fputs("    --set-groups or -G: set additional groups as in /etc/group\n", stderr);
	exit(0);
}

char *root;
int groups;

static struct option const opts[] =
{
	{ "root",  required_argument, NULL, 'r' },
	{ "caps",  required_argument, NULL, 'c' },
	{ "user",  required_argument, NULL, 'u' },
	{ "set-groups", no_argument,  NULL, 'G' },
	{ "help",  no_argument,       NULL, 'h' },
	{ NULL, 0, NULL, 0 }
};

#define OPTS "r:c:u:Gh"

int main(int argc, char *const*argv, char *const*envp)
{
	int optc;
	struct passwd *pw;
	cap_t caps, tcaps;
	char tmp[1024];

	pw = NULL;
	caps = NULL;
	tcaps = NULL;
	root = NULL;
	groups = 0;

	if (getuid() && !geteuid())
		err_die("security violation: setuid root is prohibited");

	while ((optc=getopt_long(argc, argv, OPTS, opts, NULL))!=-1) {
		switch (optc) {
		case 'r':
			root = optarg;
			break;
		case 'c':
			if (strlen(optarg) >980)
				err_die("capability list is way too long");
			strcpy(tmp, optarg);
			caps = cap_from_text(tmp);
			if (!caps)
				err_die2("requested capabilities were not recognized: ", tmp);
			strncat(tmp, " cap_setuid,cap_setgid,cap_sys_chroot+ep", 1024);
			tcaps = cap_from_text(tmp);
			if (!tcaps)
				err_die2("requested temporary capabilities were not recognized: ", tmp);
			break;
		case 'u':
			pw = getpwnam((const char*)optarg);
			if (!pw)
				err_die2("unknown account: ", optarg);
			break;
		case 'h':
			help_die();
		case 'G':
			groups = 1;
			break;
		case ':':
		case '?':
		default:
			err_die("parameter error");
		}
	}		
	if (optind >= argc)
		err_die("nothing to run");

// SET_CAPS
	if (tcaps) {
		if (cap_set_proc(tcaps))
			err_die("unable to set temporary capabilities");
		if (pw && pw->pw_uid)
			if (prctl(PR_SET_KEEPCAPS, 1))
				err_die("cannot keep capabilities");
	}

// CHROOT
	if (root) {
		if (chdir(root) == -1)
			err_die2("unable to chdir to ", root);
		if (chroot(".") == -1)
			err_die2("unable to chroot to ", root);
	}

// SET USER
	if (pw) {
		if (groups) {
			if (initgroups(pw->pw_name, pw->pw_gid) == -1)
				err_die("cannot init groups");
		} else {
			gid_t grp[2];
			grp[0] = pw->pw_gid;
			grp[1] = 70;
			if (setgroups(1, grp) == -1)
				err_die("cannot set group");
		}
		setgid(pw->pw_gid);
		setuid(pw->pw_uid);
	}

// SET_CAPS
	if (caps) {
		if (cap_set_proc(caps))
			err_die("unable to set capabilities");
	}

// EXEC
	execve(argv[optind], argv+optind, envp);
}
